import asyncio

from subito.operations import operation


class Subscription(operation.Operation):
    def __init__(self, document: str):
        """ Initiates a subscription monitor. \nIt accepts a document argument which represents a path to a file or a string representing graphql subscription and a handler function to process messages """
        super().__init__(document)
        self._handler = None

    async def consume(self, session):
        """ Subscribe to a subscription, and process it's messages """
        async for message in session.subscribe(self.ast):
            await self.process(message)

    async def process(self, data):
        """ Send to a service for further processing """
        return await self.handler(data)

    @property
    def handler(self):
        return self._handler

    @handler.setter
    def handler(self, handler):
        self._handler = handler



from asyncio import gather

from gql import Client, WebsocketsTransport

from subito.operations.subscription import Subscription
from subito.operations.operation import Operation


class Subito:
    session = None
    subscriptions = []

    def __init__(self, url: str):
        self.url = url
        self.options = {
            "transport": WebsocketsTransport(url=self.url),
            "fetch_schema_from_transport": True,
        }

    async def run(self):
        async with Client(**self.options) as session:
            Subito.session = session
            await gather(*(s.consume(session) for s in Subito.subscriptions))

    @staticmethod
    def subscribe(document: str):
        Subito.subscriptions.append(Subscription(document))

        def inner(handler):
            subscription = Subito.subscriptions[-1]
            subscription.handler = handler

        return inner

    # # TODO: What if it was made to be a @staticmethod
    # def subscribe(self, document: str):
    #     Subito.subscriptions.append(Subscription(document))

    #     def inner(handler):
    #         subscription = Subito.subscriptions[-1]
    #         subscription.handler = handler

    #         async def wrapper(*args, **kwargs):
    #             return handler

    #         return wrapper

    #     return inner

    @staticmethod
    async def execute(document: str, variables: dict = {}) -> dict:
        operation = Operation(document, variables)
        result = await operation.execute(Subito.session)
        return result


import logging
from os.path import isfile

from gql import gql
from graphql import GraphQLSyntaxError, DocumentNode


class Operation:
    def __init__(self, document: str, variables: dict = {}):
        """ Defines an operation which parses and validates passed document """
        try:
            self.logger = logging.getLogger("__main__")
            self.document = document
            if isfile(self.document):
                with open(document) as f:
                    self.ast = gql(f.read())
            else:
                self.ast = gql(self.document)
            (self.definition,) = self.ast.definitions

            if self.definition.kind != "operation_definition":
                raise TypeError("document does not contain an an operation")

            self._variables = variables

        except GraphQLSyntaxError as e:
            raise TypeError("document has syntax errors")

    @property
    def name(self):
        """ Get operation name  """
        (selection,) = self.definition.selection_set.selections
        return selection.name.value

    async def execute(self, session):
        """ Send to a service for further processing """
        return await session.execute(self.ast, variable_values=self.variables)

    @property
    def variables(self):
        return self._variables

    @variables.setter
    def variables(self, variables):
        self._variables = variables